/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.wso2.registry.Artifact;
import org.wso2.registry.RegistryConstants;
import org.wso2.registry.jdbc.DatabaseConstants;

import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

public class VersionedArtifactDAO {

    /**
     * Returns the given version of the artifact. If artifact is a collection all child paths will
     * be added as versioned paths in the form /a/b?v=12
     *
     * @throws SQLException
     * @return Artifact for the given version number
     * @param path : path of the artifact
     */
    public Artifact get(String path, long versionNumber, Connection conn) throws SQLException {

        Artifact artifact = null;

        ResultSet result;

        // get the artifact data from the artifact table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            artifact = new Artifact();

            artifact.setId(result.getLong(DatabaseConstants.AID_FIELD));
            artifact.setPath(path + "?v=" + versionNumber);
            artifact.setParentPath(getParentPath(artifact.getId(), versionNumber, conn));
            artifact.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            artifact.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            artifact.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            artifact.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            artifact.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            artifact.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (artifact == null) {
            return null;
        }

        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, artifact.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            artifact.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            artifact.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            artifact.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get artifact chidren

        if (artifact.isDirectory()) {

            String[] children = getChildPaths(artifact.getId(), versionNumber, conn);
            artifact.setContent(children);
        }

        // get artifact properties

        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, artifact.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        artifact.setProperties(props);

        return artifact;
    }

    public Artifact getArtifactByID(long artifactID, long versionNumber, Connection conn) throws SQLException {

        Artifact artifact = null;

        ResultSet result;

        // get the artifact data from the artifact table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.AID=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setLong(1, artifactID);

        result = statement.executeQuery();

        if (result.next()) {
            artifact = new Artifact();

            artifact.setId(result.getLong(DatabaseConstants.AID_FIELD));
            artifact.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            artifact.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            artifact.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            artifact.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            artifact.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            artifact.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            artifact.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (artifact == null) {
            return null;
        }

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(artifactID, conn);
        }
        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, artifact.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            artifact.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            artifact.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            artifact.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get artifact chidren

        if (artifact.isDirectory()) {

            String[] children = getChildPaths(artifact.getId(), versionNumber, conn);
            artifact.setContent(children);
        }

        // get artifact properties

        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, artifact.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        artifact.setProperties(props);

        return artifact;
    }

    /**
     * Returns the latest (current) version of an artifact. If artifact is a collection, all child
     * paths will be added as latest paths in the form /a/b
     *
     * @param path : path of the artifacy
     * @param conn :Connection to DB
     * @return  : Result artifcat
     * @throws SQLException : if somthing went wrong
     */
    public Artifact getLatestVersion(String path, Connection conn) throws SQLException {

        Artifact artifact = null;

        ResultSet result;

        // get the artifact data from the artifact table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            artifact = new Artifact();

            artifact.setId(result.getLong(DatabaseConstants.AID_FIELD));
            artifact.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            artifact.setParentPath(getParentPath(artifact.getPath()));
            artifact.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            artifact.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            artifact.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            artifact.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            artifact.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            artifact.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (artifact == null) {
            return null;
        }

        long latestVersionNumber = getLatestVerisonNumber(artifact.getId(), conn);

        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, artifact.getId());
        versionQuery.setLong(2, latestVersionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            artifact.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            artifact.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            artifact.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }
        // get artifact chidren
        if (artifact.isDirectory()) {

            String[] children = getLatestChildPaths(artifact.getId(), latestVersionNumber, conn);
            artifact.setContent(children);
        }

        // get artifact properties
        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, artifact.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        artifact.setProperties(props);
        return artifact;
    }

    public String[] getChildPaths(long artifcatID, long versionNumber,
                                  Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            long childVersion = result.getLong("DVN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH") + "?v=" + childVersion;
                dependencyPathList.add(path);
            }
        }
        return (String[]) (dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    public String getParentPath(long artifcatID, long versionNumber,
                                Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(artifcatID, connection);
        }
        if (versionNumber == -1) {
            return null;
        }
        String selectSql = "SELECT AID , VN FROM DEPENDENCY WHERE DAID=? AND DVN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        String parentPath = "";

        //to see whether the patent node is there in the table , if not need to add that
        if (result.next()) {
            long aid = result.getLong("AID");
            long parentVersionNumber = result.getLong("VN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, aid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                parentPath = path + "?v=" + parentVersionNumber;
            }
        }
        return parentPath;
    }

    public String[] getLatestChildPaths(long artifcatID, long versionNumber,
                                        Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                dependencyPathList.add(path);
            }
        }
        return (String[]) (dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    public List getChildren(long artifcatID, long versionNumber,
                            Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(artifcatID, connection);
        }

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList childList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {

            long daid = result.getLong("DAID");
            String SQL = "SELECT AID, PATH, DIRECTORY FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);

            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                Artifact child = new Artifact();
                child.setId(pathResult.getLong(DatabaseConstants.AID_FIELD));
                child.setPath(pathResult.getString(DatabaseConstants.PATH_FIELD));
                child.setDirectory(pathResult.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
                childList.add(child);
            }
        }
        return childList;
    }

    /**
     * Adds the given artifact to the database.
     *
     * @param path Path of the new artifact
     * @param artifact New artifact
     * @param conn : Connection to Db
     */
    public void add (String path, Artifact artifact, Connection conn) throws SQLException {

        // first add the artifact data into the artifacts table

        String sql = "INSERT INTO ARTIFACTS (PATH, MEDIA_TYPE, DIRECTORY, STATE, CREATED_TIME, AUTHOR, LAST_UPDATER, LAST_UPDATED_TIME,  DESCRIPTION, CONTENT)" +
                "VALUES (?,?,?,?,?,?,?,?,?,?)";

        long now = System.currentTimeMillis();

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, artifact.getPath());
        s.setString(2, artifact.getMediaType());
        s.setInt(3, artifact.isDirectory()? 1: 0);
        s.setInt(4, RegistryConstants.ACTIVE_STATE);
        s.setTimestamp(5, new Timestamp(now));
        s.setString(6, artifact.getAuthorUserName());
        s.setString(7, artifact.getLastUpdaterUserName());
        s.setTimestamp(8, new Timestamp(now));
        s.setString(9, artifact.getDescription());

        if (artifact.getContent() != null && artifact.getContent() instanceof byte[]) {
            s.setBytes(10, (byte[]) artifact.getContent());
        } else {
            s.setBytes(10, null);
        }

        s.executeUpdate();


        long newArtifactID = this.getArtifactID(artifact.getPath(), conn);

        // now add artifact properties to the properties table
        addProperties(newArtifactID, artifact.getProperties(), conn);
    }

    public void update(String path, Artifact artifact, Connection conn) throws SQLException{

        // first update the artifacts table

        String sql = "UPDATE ARTIFACTS SET MEDIA_TYPE=?, DIRECTORY=?, STATE=?, AUTHOR=?, LAST_UPDATER=?, LAST_UPDATED_TIME=?, DESCRIPTION=?, CONTENT=? WHERE PATH=?";

        long now = System.currentTimeMillis();

        PreparedStatement s = conn.prepareStatement(sql);

        s.setString(1, artifact.getMediaType());
        s.setInt(2, artifact.isDirectory()? 1: 0);
        s.setInt(3, artifact.getState());
        s.setString(4, artifact.getAuthorUserName());
        s.setString(5, artifact.getLastUpdaterUserName());
        s.setTimestamp(6, new Timestamp(now));
        s.setString(7, artifact.getDescription());

        if (artifact.getContent() != null && artifact.getContent() instanceof byte[]) {
            s.setBytes(8, (byte[]) artifact.getContent());
        } else {
            s.setBytes(8, null);
        }

        s.setString(9, path);

        s.executeUpdate();

        // remove existing properties for the updated artfact
        removeProperties(artifact.getId(), conn);

        // now add the updated properties to the properties table
        addProperties(artifact.getId(), artifact.getProperties(), conn);
    }

    public void markDeleted(String path, Connection conn) throws SQLException {

        String sql = "UPDATE ARTIFACTS SET STATE=? WHERE PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setInt(1, RegistryConstants.DELETED_STATE);
        s.setString(2, path);
        s.executeUpdate();
    }

    public void addProperties(long artifactID, Properties props, Connection conn)
            throws SQLException {

        String propSQL = "INSERT INTO PROPERTIES (AID, PKEY, PVALUE) VALUES (?,?,?)";
        PreparedStatement s2 = conn.prepareStatement(propSQL);

        if (props != null) {

            Iterator i = props.keySet().iterator();
            while (i.hasNext()) {
                String key = (String) i.next();
                String value = (String) props.get(key);

                s2.setLong(1, artifactID);
                s2.setString(2, key);
                s2.setString(3, value);

                s2.executeUpdate();
                s2.clearParameters();
            }
        }
    }

    public void removeProperties(long artifactID, Connection conn) throws SQLException {

        String removePropsSQL = "DELETE FROM PROPERTIES WHERE AID=?";
        PreparedStatement s2 = conn.prepareStatement(removePropsSQL);
        s2.setLong(1, artifactID);
        s2.executeUpdate();
    }

    public boolean artifactExist(String path, Connection conn) throws SQLException {

        String sql = "SELECT COUNT(A.AID) FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        long count = 0;
        if (result.next()) {
            count = result.getLong(1);
        }

        return count == 1;
    }

    public long getArtifactID(String path, Connection conn) throws SQLException {

        String sql = "SELECT A.AID FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        if (result.next()) {
            return result.getLong(DatabaseConstants.AID_FIELD);
        }

        return -1;
    }

    public void addResourceVersion(Artifact artifact, Connection connection) throws SQLException {

        if (artifact.getLastModified() == null) {
            artifact.setLastModified(new java.util.Date());
        }

        long versionNumber = insertToVersionTable(artifact, connection);

        //updating max version table
        updateDependency(artifact, versionNumber, false, connection);
    }

    private long insertToVersionTable(Artifact artifact, Connection connection)
            throws SQLException {

        String sql = "INSERT INTO VERSIONS (AID, VN ,CONTENT , AUTHOR, UPDATED_TIME) VALUES (?,?,?,?,?)";
        PreparedStatement statement = connection.prepareStatement(sql);
        statement.setLong(1, artifact.getId());

        // Need to getthe old version and increment that
        // Selecet VID from latest table ,
        // if it is there then find from that  , if it is not there then we need ti start from 0
        long versionNumber = getLatestVerisonNumber(artifact.getId(), connection);
        if (versionNumber == -1) {
            versionNumber = 1;
        } else {
            //incrementing the version number
            versionNumber = versionNumber + 1;
        }
        statement.setLong(2, versionNumber);

        Object content = artifact.getContent();
        if (content != null && content instanceof byte[]) {
            statement.setBytes(3, (byte[]) content);
        } else {
            statement.setBytes(3, null);
        }

        statement.setString(4, artifact.getLastUpdaterUserName());
        statement.setTimestamp(5, new Timestamp(System.currentTimeMillis()));
        statement.executeUpdate();

        return versionNumber;
    }

    private void updateDependency(Artifact artifact,
                                  long versionNumber,
                                  boolean delete, Connection connection) throws SQLException {

        // updating dependency table , so that we can rollback to a particular version
        // updating for the current artifact
        long parentNextVersionNumber =
                updateDependencyTable(artifact, versionNumber, delete, connection);

        // a/b/c/r1.wsdl
        // Next a/b/c
        // Next a/b
        // Next a
        if (parentNextVersionNumber != -1) {

            // updating its parent
            String path = artifact.getPath();
            while (path != null && path.indexOf("/") > -1) {

                int lastIndex = path.lastIndexOf("/");
                String parentPath = path.substring(0, lastIndex);
                Artifact parentArtifact = getLatestVersion(parentPath, connection);

                if (parentArtifact != null) {
                    path = parentArtifact.getPath();
                    parentNextVersionNumber = updateDependencyTable(
                            parentArtifact, parentNextVersionNumber, false, connection);

                    if (parentNextVersionNumber == -1) {
                        break;
                    }
                } else {
                    break;
                }
            }
        }
    }

    /**
     * @param artifact      : Artifcat we need to update itslef and its immediate parent
     * @param versionNumber : Version number of the artifcat
     * @param delete        : delete or adding
     * @param connection    : DB connetion
     * @return : Will retuen the artifact version of the parent artifact
     * @throws SQLException : if something went wrong
     */
    private long updateDependencyTable(Artifact artifact,
                                       long versionNumber,
                                       boolean delete, Connection connection) throws SQLException {
        /**
         *        D0
         *    D1      D2
         *  D3  R1    R2
         *
         *  First fine D1 V0
         */
        // Checking whether the path is null or top most directory
        // Checking whether artifact is root node or not

        String path = artifact.getPath();

        // Checking whether the path is null or top most directory
        // if (path != null && path.indexOf("/") > -1) {
        if (path != null && !path.equals(RegistryConstants.ROOT_PATH)) {

            int lastIndex = path.lastIndexOf("/");
            String parentPath = RegistryConstants.ROOT_PATH;
            if (lastIndex != 0) {
                parentPath = path.substring(0, lastIndex);
            }

            if ("".equals(parentPath)) {
                //TODO : Need to think the logic here

            } else {

                long parentId = getArtifactID(parentPath, connection);

                if (artifact.getLastModified() == null) {
                    artifact.setLastModified(new java.util.Date());
                }
                // Need to get the parent's parent and update the and update the

                long artifactversionNumber = getLatestVerisonNumber(parentId, connection);

                long nextVersionNumber = createNextVersion(parentId,
                        artifact.getLastUpdaterUserName(),
                        artifact.getLastModified().getTime(), connection);

                String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
                PreparedStatement s = connection.prepareStatement(selectSql);
                s.setLong(1, parentId);
                s.setLong(2, artifactversionNumber);
                ResultSet result = s.executeQuery();

                ArrayList dependeyMap = new ArrayList();

                // to see whether the parent node is there in the table , if not need to add that
                while (result.next()) {

                    long daid = result.getLong("DAID");

                    if (daid != artifact.getId()) {
                        long dvn = result.getLong("DVN");
                        PreparedStatement ps = getStatementForDependencyTable(connection,
                                parentId, nextVersionNumber, daid, dvn);
                        dependeyMap.add(ps);
                    }
                }

                if (!delete) {
                    PreparedStatement ds = getStatementForDependencyTable(connection,
                            parentId, nextVersionNumber, artifact.getId(), versionNumber);
                    dependeyMap.add(ds);
                }

                for (int i = 0; i < dependeyMap.size(); i++) {
                    PreparedStatement preparedStatement = (PreparedStatement) dependeyMap.get(i);
                    preparedStatement.executeUpdate();
                }

                return nextVersionNumber;
            }

        } else {
            // TODO , I think we do not need to add the entry here , since when adding a child node,
            // if the parent is not there then we will add that
        }

        return -1;
    }

    public long getLatestVerisonNumber(long artifactId, Connection connection) throws SQLException {

        String sql = "SELECT MAX(VN) FROM VERSIONS WHERE AID=?";
        PreparedStatement s = connection.prepareStatement(sql);
        s.setLong(1, artifactId);
        ResultSet result = s.executeQuery();

        if (result.next()) {
            return result.getLong(1);
        }

        return -1;
    }

    public String[] getVersionNumbers(long artifactId, Connection connection) throws SQLException {

        String SQL = "SELECT VN FROM VERSIONS WHERE AID=?";
        PreparedStatement s = connection.prepareStatement(SQL);
        s.setLong(1, artifactId);
        ResultSet result = s.executeQuery();
        ArrayList versionNumbers = new ArrayList();
        while (result.next()) {
            versionNumbers.add("" + result.getLong("VN"));
        }
        return (String[]) versionNumbers.toArray(new String[versionNumbers.size()]);
    }

    public void restore(long artifactId, long toVersion, Connection connection) throws SQLException {

        Artifact oldArtifact = getArtifactByID(artifactId, toVersion, connection);

        String plainPath = oldArtifact.getPath();
        if (oldArtifact.getPath().indexOf("?v=") != -1) {
            String[] parts = oldArtifact.getPath().split("\\?v=");
            plainPath = parts[0];
        }

        update(plainPath, oldArtifact, connection);
        addResourceVersion(oldArtifact, connection);

        if (oldArtifact.isDirectory()) {

            String selectChildSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
            PreparedStatement selectps = connection.prepareStatement(selectChildSql);
            selectps.setLong(1, artifactId);
            selectps.setLong(2, toVersion);
            ResultSet childResult = selectps.executeQuery();
            ArrayList dependencyList = new ArrayList();

            while (childResult.next()) {

                long daid = childResult.getLong("DAID");
                long dvn = childResult.getLong("DVN");
                long versionNumber = getLatestVerisonNumber(artifactId, connection);
                PreparedStatement ps = getStatementForDependencyTable(connection,
                        artifactId, versionNumber, daid, dvn);
                dependencyList.add(ps);
            }

            for (int i = 0; i < dependencyList.size(); i++) {
                PreparedStatement preparedStatement = (PreparedStatement) dependencyList.get(i);
                preparedStatement.executeUpdate();
            }
        }
    }

    /**
     * To restoreVersion or rollback artifcat to a given version number
     *
     * @param artifactId : Artifact ID
     * @param connection : Connection to DB
     * @param toVersion  : to which version we need to restoreVersion
     * @throws SQLException : If something went wrong
     */
    public void restore2(long artifactId, long toVersion, Connection connection) throws SQLException {

        String SQL = "SELECT CONTENT, AUTHOR, UPDATED_TIME FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement statement = connection.prepareStatement(SQL);
        statement.setLong(1, artifactId);
        statement.setLong(2, toVersion);

        ResultSet result = statement.executeQuery();
        long versionNumber = 0;
        if (result.next()) {
            Artifact artifactToRestore = new Artifact();
            artifactToRestore.setContent(result.getObject(1));
            artifactToRestore.setAuthorUserName(result.getString(2));
            artifactToRestore.setLastModified(result.getDate(3));
            versionNumber = insertToVersionTable(artifactToRestore, connection);
        }

        Artifact artifact = getArtifactByID(artifactId, toVersion, connection);

        // updating max version table
        updateDependency(artifact, versionNumber, false, connection);
        if (artifact.isDirectory()) {
            String selectChildSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
            PreparedStatement selectps = connection.prepareStatement(selectChildSql);
            selectps.setLong(1, artifactId);
            selectps.setLong(2, toVersion);
            ResultSet childResult = selectps.executeQuery();
            ArrayList dependencyList = new ArrayList();
            while (childResult.next()) {
                long daid = childResult.getLong("DAID");
                long dvn = childResult.getLong("DVN");
                PreparedStatement ps = getStatementForDependencyTable(connection,
                        artifactId, versionNumber, daid, dvn);
                dependencyList.add(ps);
            }

            for (int i = 0; i < dependencyList.size(); i++) {
                PreparedStatement preparedStatement = (PreparedStatement) dependencyList.get(i);
                preparedStatement.executeUpdate();
            }
        }

    }

    /**
     * To delete a artifact , what actualy happen here is that update the version table with the new data
     * and artifact will not actualy removed from the table
     *
     * @param artifact   : Artifact to be deleted
     * @param connection : Connection to the DB
     * @throws java.sql.SQLException : if something went wrong
     */
    public void deleteResource(Artifact artifact,
                               Connection connection) throws SQLException {
        long versionNumber = getLatestVerisonNumber(artifact.getId(), connection);
        updateDependency(artifact, versionNumber, true, connection);
    }

    private long createNextVersion(long artifactID,
                                   String lastUpdaterUserName,
                                   long lastModifiedTime,
                                   Connection connection) throws SQLException {

        long maxVersion = getLatestVerisonNumber(artifactID, connection);
        if (maxVersion == -1) {
            maxVersion = 0;
        }

        // here maxVersion is always >0
        maxVersion = maxVersion + 1;

        Artifact artifact = new Artifact();
        artifact.setId(artifactID);
        artifact.setLastUpdaterUserName(lastUpdaterUserName);
        artifact.setLastModified(new Date(lastModifiedTime));
        insertToVersionTable(artifact, connection);

        return maxVersion;
    }

    private PreparedStatement getStatementForDependencyTable(Connection connection,
                                                             long parentId,
                                                             long nextVersionNumber,
                                                             long daid,
                                                             long dvn) throws SQLException {

        String SQL = "INSERT INTO DEPENDENCY (AID, VN , DAID , DVN) VALUES (?,? ,?,?)";
        PreparedStatement ps = connection.prepareStatement(SQL);
        ps.setLong(1, parentId);
        ps.setLong(2, nextVersionNumber);
        ps.setLong(3, daid);
        ps.setLong(4, dvn);
        System.out.println("Adding dependency: " + parentId + " " + nextVersionNumber + " " + daid + " " + dvn);

        return ps;
    }

    private String getParentPath(String path) {

        String[] parts = path.split(RegistryConstants.PATH_SEPARATOR);
        String parentPath = "";

        if (parts.length > 2) {
            for (int i = 1; i < parts.length - 1; i++) {
                parentPath = parentPath + "/" + parts[i];

            }
        } else {
            parentPath = "/";
        }

        return parentPath;
    }
}
